from .augmentations.spectrogram_transforms import SpecFrequencyMask, SpecChannelShuffle
from .augmentations.transforms import (
    AddBackgroundNoise,
    AddGaussianNoise,
    AddGaussianSNR,
    AddImpulseResponse,
    AddShortNoises,
    ClippingDistortion,
    FrequencyMask,
    Gain,
    LoudnessNormalization,
    Mp3Compression,
    Normalize,
    PitchShift,
    PolarityInversion,
    Resample,
    Shift,
    TimeMask,
    TimeStretch,
    Trim,
)
from .core.composition import Compose, SpecCompose

__version__ = "0.16.0"



import os
from pathlib import Path

import numpy as np

AUDIO_FILENAME_ENDINGS = (".aiff", ".flac", ".m4a", ".mp3", ".ogg", ".opus", ".wav")


def get_file_paths(
    root_path, filename_endings=AUDIO_FILENAME_ENDINGS, traverse_subdirectories=True
):
    """Return a list of paths to all files with the given filename extensions in a directory.
    Also traverses subdirectories by default.
    """
    file_paths = []

    for root, dirs, filenames in os.walk(root_path):
        filenames = sorted(filenames)
        for filename in filenames:
            input_path = os.path.abspath(root)
            file_path = os.path.join(input_path, filename)

            if filename.lower().endswith(filename_endings):
                file_paths.append(Path(file_path))
        if not traverse_subdirectories:
            # prevent descending into subfolders
            break

    return file_paths


def calculate_rms(samples):
    """Given a numpy array of audio samples, return its Root Mean Square (RMS)."""
    return np.sqrt(np.mean(np.square(samples), axis=-1))


def calculate_desired_noise_rms(clean_rms, snr):
    """
    Given the Root Mean Square (RMS) of a clean sound and a desired signal-to-noise ratio (SNR),
    calculate the desired RMS of a noise sound to be mixed in.

    Based on https://github.com/Sato-Kunihiko/audio-SNR/blob/8d2c933b6c0afe6f1203251f4877e7a1068a6130/create_mixed_audio_file.py#L20
    :param clean_rms: Root Mean Square (RMS) - a value between 0.0 and 1.0
    :param snr: Signal-to-Noise (SNR) Ratio in dB - typically somewhere between -20 and 60
    :return:
    """
    a = float(snr) / 20
    noise_rms = clean_rms / (10 ** a)
    return noise_rms


def convert_decibels_to_amplitude_ratio(decibels):
    return 10 ** (decibels / 20)


def is_waveform_multichannel(samples):
    """
    Return bool that answers the question: Is the given ndarray a multichannel waveform or not?

    :param samples: numpy ndarray
    :return:
    """
    return len(samples.shape) > 1


def is_spectrogram_multichannel(spectrogram):
    """
    Return bool that answers the question: Is the given ndarray a multichannel spectrogram?

    :param samples: numpy ndarray
    :return:
    """
    return len(spectrogram.shape) > 2 and spectrogram.shape[-1] > 1


def convert_float_samples_to_int16(y):
    """Convert floating-point numpy array of audio samples to int16."""
    if not issubclass(y.dtype.type, np.floating):
        raise ValueError("input samples not floating-point")
    return (y * np.iinfo(np.int16).max).astype(np.int16)



import random

import numpy as np

from audiomentations.core.transforms_interface import BaseSpectrogramTransform


class SpecFrequencyMask(BaseSpectrogramTransform):
    """
    Mask a set of frequencies in a spectrogram, à la Google AI SpecAugment. This type of data
    augmentation has proved to make speech recognition models more robust.

    The masked frequencies can be replaced with either the mean of the original values or a
    given constant (e.g. zero).
    """

    supports_multichannel = True

    def __init__(
        self,
        min_mask_fraction: float = 0.03,
        max_mask_fraction: float = 0.25,
        fill_mode: str = "constant",
        fill_constant: float = 0.0,
        p: float = 0.5,
    ):
        super().__init__(p)
        self.min_mask_fraction = min_mask_fraction
        self.max_mask_fraction = max_mask_fraction
        assert fill_mode in ("mean", "constant")
        self.fill_mode = fill_mode
        self.fill_constant = fill_constant

    def randomize_parameters(self, magnitude_spectrogram):
        super().randomize_parameters(magnitude_spectrogram)
        if self.parameters["should_apply"]:
            num_frequency_bins = magnitude_spectrogram.shape[0]
            min_frequencies_to_mask = int(
                round(self.min_mask_fraction * num_frequency_bins)
            )
            max_frequencies_to_mask = int(
                round(self.max_mask_fraction * num_frequency_bins)
            )
            num_frequencies_to_mask = random.randint(
                min_frequencies_to_mask, max_frequencies_to_mask
            )
            self.parameters["start_frequency_index"] = random.randint(
                0, num_frequency_bins - num_frequencies_to_mask
            )
            self.parameters["end_frequency_index"] = (
                self.parameters["start_frequency_index"] + num_frequencies_to_mask
            )

    def apply(self, magnitude_spectrogram):
        if self.fill_mode == "mean":
            fill_value = np.mean(
                magnitude_spectrogram[
                    self.parameters["start_frequency_index"] : self.parameters[
                        "end_frequency_index"
                    ]
                ]
            )
        else:
            # self.fill_mode == "constant"
            fill_value = self.fill_constant
        magnitude_spectrogram = magnitude_spectrogram.copy()
        magnitude_spectrogram[
            self.parameters["start_frequency_index"] : self.parameters[
                "end_frequency_index"
            ]
        ] = fill_value
        return magnitude_spectrogram


class SpecChannelShuffle(BaseSpectrogramTransform):
    """
    Shuffle the channels of a multichannel spectrogram (channels last).
    This can help combat positional bias.
    """
    supports_multichannel = True
    supports_mono = False

    def randomize_parameters(self, magnitude_spectrogram):
        super().randomize_parameters(magnitude_spectrogram)
        if self.parameters["should_apply"]:
            self.parameters["shuffled_channel_indexes"] = list(range(magnitude_spectrogram.shape[-1]))
            random.shuffle(self.parameters["shuffled_channel_indexes"])

    def apply(self, magnitude_spectrogram):
        return magnitude_spectrogram[..., self.parameters["shuffled_channel_indexes"]]



import random


class BaseCompose:
    def __init__(self, transforms, p=1.0, shuffle=False):
        self.transforms = transforms
        self.p = p
        self.shuffle = shuffle

        name_list = []
        for transform in self.transforms:
            name_list.append(type(transform).__name__)
        self.__name__ = "_".join(name_list)

    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    def randomize_parameters(self, *args, **kwargs):
        """
        Randomize and define parameters of every transform in composition.
        """
        raise NotImplementedError

    def freeze_parameters(self):
        """
        Mark all parameters as frozen, i.e. do not randomize them for each call. This can be
        useful if you want to apply an effect chain with the exact same parameters to multiple
        sounds.
        """
        for transform in self.transforms:
            transform.freeze_parameters()

    def unfreeze_parameters(self):
        """
        Unmark all parameters as frozen, i.e. let them be randomized for each call.
        """
        for transform in self.transforms:
            transform.unfreeze_parameters()


class Compose(BaseCompose):
    # TODO: Name can change to WaveformCompose
    def __init__(self, transforms, p=1.0, shuffle=False):
        super(Compose, self).__init__(transforms, p, shuffle)

    def __call__(self, samples, sample_rate):
        transforms = self.transforms.copy()
        if random.random() < self.p:
            if self.shuffle:
                random.shuffle(transforms)
            for transform in transforms:
                samples = transform(samples, sample_rate)

        return samples

    def randomize_parameters(self, samples, sample_rate):
        for transform in self.transforms:
            transform.randomize_parameters(samples, sample_rate)


class SpecCompose(BaseCompose):
    def __init__(self, transforms, p=1.0, shuffle=False):
        super().__init__(transforms, p, shuffle)

    def __call__(self, magnitude_spectrogram):
        transforms = self.transforms.copy()
        if random.random() < self.p:
            if self.shuffle:
                random.shuffle(transforms)
            for transform in transforms:
                magnitude_spectrogram = transform(magnitude_spectrogram)

        return magnitude_spectrogram

    def randomize_parameters(self, magnitude_spectrogram):
        for transform in self.transforms:
            transform.randomize_parameters(magnitude_spectrogram)


import codecs
import os
import re

from setuptools import setup, find_packages

with open("README.md", "r") as readme_file:
    long_description = readme_file.read()

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="audiomentations",
    version=find_version("audiomentations", "__init__.py"),
    author="Iver Jordal",
    description="A Python library for audio data augmentation. Inspired by albumentations."
    " Useful for machine learning.",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/iver56/audiomentations",
    packages=find_packages(exclude=["demo", "tests"]),
    install_requires=["numpy>=1.13.0", "librosa>=0.6.1,<=0.8.0", "scipy>=1.0.0,<1.6.0"],
    extras_require={
        "extras": [
            "lameenc>=1.2.0,<2",
            "pydub>=0.22.0,<1",
            "pyloudnorm>=0.1.0",
            "wavio>=0.0.4",
        ]
    },
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Multimedia",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



import random
import warnings

from audiomentations.core.utils import (
    is_waveform_multichannel,
    is_spectrogram_multichannel,
)


class MultichannelAudioNotSupportedException(Exception):
    pass


class MonoAudioNotSupportedException(Exception):
    pass


class BaseTransform:
    supports_mono = True
    supports_multichannel = False

    def __init__(self, p=0.5):
        assert 0 <= p <= 1
        self.p = p
        self.parameters = {"should_apply": None}
        self.are_parameters_frozen = False

    def serialize_parameters(self):
        """Return the parameters as a JSON-serializable dict."""
        return self.parameters

    def freeze_parameters(self):
        """
        Mark all parameters as frozen, i.e. do not randomize them for each call. This can be
        useful if you want to apply an effect with the exact same parameters to multiple sounds.
        """
        self.are_parameters_frozen = True

    def unfreeze_parameters(self):
        """
        Unmark all parameters as frozen, i.e. let them be randomized for each call.
        """
        self.are_parameters_frozen = False


class BaseWaveformTransform(BaseTransform):
    def apply(self, samples, sample_rate):
        raise NotImplementedError

    def is_multichannel(self, samples):
        return is_waveform_multichannel(samples)

    def __call__(self, samples, sample_rate):
        if not self.are_parameters_frozen:
            self.randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"] and len(samples) > 0:
            if self.is_multichannel(samples):
                if samples.shape[0] > samples.shape[1]:
                    warnings.warn(
                        "Multichannel audio must have channels first, not channels last. In"
                        " other words, the shape must be (channels, samples), not"
                        " (samples, channels)"
                    )
                if not self.supports_multichannel:
                    raise MultichannelAudioNotSupportedException(
                        "{} only supports mono audio, not multichannel audio".format(
                            self.__class__.__name__
                        )
                    )
            elif not self.supports_mono:
                raise MonoAudioNotSupportedException(
                    "{} only supports multichannel audio, not mono audio".format(
                        self.__class__.__name__
                    )
                )
            return self.apply(samples, sample_rate)
        return samples

    def randomize_parameters(self, samples, sample_rate):
        self.parameters["should_apply"] = random.random() < self.p


class BaseSpectrogramTransform(BaseTransform):
    def apply(self, magnitude_spectrogram):
        raise NotImplementedError

    def is_multichannel(self, samples):
        return is_spectrogram_multichannel(samples)

    def __call__(self, magnitude_spectrogram):
        if not self.are_parameters_frozen:
            self.randomize_parameters(magnitude_spectrogram)
        if (
            self.parameters["should_apply"]
            and magnitude_spectrogram.shape[0] > 0
            and magnitude_spectrogram.shape[1] > 0
        ):
            if self.is_multichannel(magnitude_spectrogram):
                """
                if magnitude_spectrogram.shape[0] > magnitude_spectrogram.shape[1]:
                    warnings.warn(
                        "Multichannel audio must have channels first, not channels last"
                    )
                """
                if not self.supports_multichannel:
                    raise MultichannelAudioNotSupportedException(
                        "{} only supports mono audio, not multichannel audio".format(
                            self.__class__.__name__
                        )
                    )
            elif not self.supports_mono:
                raise MonoAudioNotSupportedException(
                    "{} only supports multichannel audio, not mono audio".format(
                        self.__class__.__name__
                    )
                )

            return self.apply(magnitude_spectrogram)
        return magnitude_spectrogram

    def randomize_parameters(self, magnitude_spectrogram):
        self.parameters["should_apply"] = random.random() < self.p



import warnings
from pathlib import Path

import librosa
import numpy as np
from scipy.io import wavfile

IS_WAVIO_INSTALLED = True
try:
    import wavio
except ImportError:
    IS_WAVIO_INSTALLED = False


def load_sound_file(file_path, sample_rate, mono=True, resample_type="auto"):
    """
    Load an audio file as a floating point time series. Audio will be automatically
    resampled to the given sample rate.

    :param file_path: str or Path instance that points to a sound file
    :param sample_rate: If not None, resample to this sample rate
    :param mono: If True, mix any multichannel data down to mono, and return a 1D array
    :param resample_type: "auto" means use "kaiser_fast" when upsampling and "kaiser_best" when
        downsampling
    """
    file_path = Path(file_path)
    if file_path.name.lower().endswith(".wav"):
        # Use scipy for loading most wav files, because scipy is fast
        try:
            return load_wav_file(
                file_path, sample_rate, mono, resample_type=resample_type
            )
        except Exception as e:
            # scipy<1.6.0 does not natively support 24-bit wavs, so we use wavio or librosa.
            if "the wav file has 24-bit data" in str(e):
                if IS_WAVIO_INSTALLED:
                    return load_wav_file_with_wavio(
                        file_path, sample_rate, mono, resample_type=resample_type
                    )
                else:
                    warnings.warn(
                        "You are loading a 24-bit wav file, and librosa is not very fast at"
                        " doing that. Install wavio for a performance boost. To install the"
                        " optional wavio dependency of audiomentations,"
                        " do `pip install audiomentations[extras]` instead of"
                        " `pip install audiomentations`"
                    )
            elif "Unknown wave file format" in str(e):
                # This can happen if the file is in MS ADPCM format
                pass
            else:
                raise e
    samples, actual_sample_rate = librosa.load(
        str(file_path), sr=None, mono=mono, dtype=np.float32
    )

    if sample_rate is not None and actual_sample_rate != sample_rate:
        if resample_type == "auto":
            resample_type = (
                "kaiser_fast" if actual_sample_rate < sample_rate else "kaiser_best"
            )
        samples = librosa.resample(
            samples, actual_sample_rate, sample_rate, res_type=resample_type
        )
        warnings.warn(
            "{} had to be resampled from {} hz to {} hz. This hurt execution time.".format(
                str(file_path), actual_sample_rate, sample_rate
            )
        )

    actual_sample_rate = actual_sample_rate if sample_rate is None else sample_rate

    if mono:
        assert len(samples.shape) == 1
    return samples, actual_sample_rate


def load_wav_file(file_path, sample_rate, mono=True, resample_type="kaiser_best"):
    """Load a wav audio file as a floating point time series. Significantly faster than
    load_sound_file."""

    actual_sample_rate, samples = wavfile.read(file_path)

    if samples.dtype != np.float32:
        if samples.dtype == np.int16:
            samples = np.true_divide(
                samples, 32768, dtype=np.float32
            )  # ends up roughly between -1 and 1
        elif samples.dtype == np.int32:
            # TODO: Add support for 24-bit and 32-bit loading in scipy>=1.6.0
            raise Exception("Unexpected data type")
        else:
            raise Exception("Unexpected data type")

    if mono and len(samples.shape) > 1:
        if samples.shape[1] == 1:
            samples = samples[:, 0]
        else:
            samples = np.mean(samples, axis=1)

    if sample_rate is not None and actual_sample_rate != sample_rate:
        if resample_type == "auto":
            resample_type = (
                "kaiser_fast" if actual_sample_rate < sample_rate else "kaiser_best"
            )

        samples = librosa.resample(
            samples, actual_sample_rate, sample_rate, res_type=resample_type
        )
        warnings.warn(
            "{} had to be resampled from {} hz to {} hz. This hurt execution time.".format(
                str(file_path), actual_sample_rate, sample_rate
            )
        )

    actual_sample_rate = actual_sample_rate if sample_rate is None else sample_rate

    return samples, actual_sample_rate


def load_wav_file_with_wavio(
    file_path, sample_rate, mono=True, resample_type="kaiser_best"
):
    """Load a 24-bit wav audio file as a floating point time series. Significantly faster than
    load_sound_file."""

    wavio_obj = wavio.read(str(file_path))
    samples = wavio_obj.data
    actual_sample_rate = wavio_obj.rate

    if samples.dtype != np.float32:
        if wavio_obj.sampwidth == 3:
            samples = np.true_divide(
                samples, 8388608, dtype=np.float32
            )  # ends up roughly between -1 and 1
        elif wavio_obj.sampwidth == 2:
            samples = np.true_divide(
                samples, 32768, dtype=np.float32
            )  # ends up roughly between -1 and 1
        else:
            raise Exception("Unknown sampwidth")

    if mono and len(samples.shape) > 1:
        if samples.shape[1] == 1:
            samples = samples[:, 0]
        else:
            samples = np.mean(samples, axis=1)

    if sample_rate is not None and actual_sample_rate != sample_rate:
        if resample_type == "auto":
            resample_type = (
                "kaiser_fast" if actual_sample_rate < sample_rate else "kaiser_best"
            )
        samples = librosa.resample(
            samples, actual_sample_rate, sample_rate, res_type=resample_type
        )
        warnings.warn(
            "{} had to be resampled from {} hz to {} hz. This hurt execution time.".format(
                str(file_path), actual_sample_rate, sample_rate
            )
        )

    actual_sample_rate = actual_sample_rate if sample_rate is None else sample_rate

    return samples, actual_sample_rate
